"""Pylama utils."""

from io import StringIO
from sys import stdin
from typing import List


def get_lines(value: str) -> List[str]:
    """Return lines from the given string."""
    return StringIO(value).readlines()


def read(filename: str) -> str:
    """Read the given filename."""
    with open(filename, encoding="utf-8") as file:
        return file.read()


def read_stdin() -> str:
    """Get value from stdin."""
    value = stdin.buffer.read()
    return value.decode("utf-8")



"""A fake linter which one never be loaded."""

from typing import Dict, Iterator, Any

import unknown_module  # noqa

from pylama.lint import Linter as Abstract


class Linter(Abstract):
    """Just a fake."""

    name = "fake"

    def run(self, path: str, **_) -> Iterator[Dict[str, Any]]:
        """Run the unknown module."""
        return unknown_module.run(path)



"""Pylama's shell support."""

import sys
import warnings
from json import dumps
from os import path as op
from os import walk
from typing import List, Optional

from .check_async import check_async
from .config import CURDIR, Namespace, parse_options, setup_logger
from .core import LOGGER, run
from .errors import Error
from .utils import read_stdin

DEFAULT_FORMAT = "{filename}:{lnum}:{col} [{etype}] {number} {message} [{source}]"
MESSAGE_FORMATS = {
    "pylint": "{filename}:{lnum}: [{etype}] {number} {message} [{source}]",
    "pycodestyle": "{filename}:{lnum}:{col} {number} {message} [{source}]",
    "parsable": DEFAULT_FORMAT,
}


def check_paths(
    paths: Optional[List[str]],
    options: Namespace,
    code: str = None,
    rootdir: str = None,
) -> List[Error]:
    """Check the given paths.

    :param rootdir: Root directory (for making relative file paths)
    :param options: Parsed pylama options (from pylama.config.parse_options)
    """
    paths = paths or options.paths
    if not paths:
        return []

    if code is None:
        candidates = []
        for path in paths or options.paths:
            if not op.exists(path):
                continue

            if not op.isdir(path):
                candidates.append(op.abspath(path))

            for root, _, files in walk(path):
                candidates += [op.relpath(op.join(root, f), CURDIR) for f in files]
    else:
        candidates = [paths[0]]

    if not candidates:
        return []

    if rootdir is None:
        path = candidates[0]
        rootdir = path if op.isdir(path) else op.dirname(path)

    candidates = [path for path in candidates if path.endswith('.py')]

    if options.concurrent:
        return check_async(candidates, code=code, options=options, rootdir=rootdir)

    errors = []
    for path in candidates:
        errors += run(path=path, code=code, rootdir=rootdir, options=options)

    return errors


def check_path(
    options: Namespace,
    rootdir: str = None,
    candidates: List[str] = None,
    code: str = None,  # noqa
) -> List[Error]:
    """Support legacy code."""
    warnings.warn(
        "pylama.main.check_path is depricated and will be removed in pylama 9",
        DeprecationWarning,
    )
    return check_paths(candidates, code=code, options=options, rootdir=rootdir)


def shell(args: List[str] = None, error: bool = True):
    """Endpoint for console.

    Parse a command arguments, configuration files and run a checkers.
    """
    if args is None:
        args = sys.argv[1:]

    options = parse_options(args)
    setup_logger(options)
    LOGGER.info(options)

    # Install VSC hook
    if options.hook:
        from .hook import install_hook  # noqa

        for path in options.paths:
            return install_hook(path)

    if options.from_stdin and not options.paths:
        LOGGER.error('--from-stdin requires a filename')
        return sys.exit(1)

    errors = check_paths(
        options.paths,
        code=read_stdin() if options.from_stdin else None,
        options=options,
        rootdir=CURDIR,
    )
    display_errors(errors, options)

    if error:
        sys.exit(int(bool(errors)))

    return errors


def display_errors(errors: List[Error], options: Namespace):
    """Format and display the given errors."""
    if options.format == "json":
        LOGGER.warning(dumps([err.to_dict() for err in errors]))

    else:
        pattern = MESSAGE_FORMATS.get(options.format, DEFAULT_FORMAT)
        for err in errors:
            LOGGER.warning(err.format(pattern))


if __name__ == "__main__":
    shell()

# pylama:ignore=F0001



"""Support Radon.

Supports stdin.
"""

from radon.complexity import add_inner_blocks
from radon.visitors import ComplexityVisitor

from pylama.context import RunContext
from pylama.lint import LinterV2 as Abstract


class Linter(Abstract):
    """Radon runner."""

    name = "radon"

    def run_check(self, context: RunContext):  # noqa  # noqa
        """Check code with Radon."""
        params = context.get_params("radon")
        complexity = params.get("complexity", 10)
        no_assert = params.get("no_assert", False)
        show_closures = params.get("show_closures", False)
        visitor = ComplexityVisitor.from_code(context.source, no_assert=no_assert)
        blocks = visitor.blocks
        if show_closures:
            blocks = add_inner_blocks(blocks)
        for block in visitor.blocks:
            if block.complexity > complexity:
                context.push(
                    lnum=block.lineno,
                    col=block.col_offset,
                    source="radon",
                    type="R",
                    number="R901",
                    text=f"{block.name} is too complex {block.complexity}",
                )



""" Don't duplicate same errors from different linters. """
from __future__ import annotations

import re
from collections import defaultdict
from typing import Any, DefaultDict, Dict, Generator, List, Set


PATTERN_NUMBER = re.compile(r'^\s*([A-Z]\d+)\s*', re.I)

DUPLICATES = {key: values for values in (

    # multiple statements on one line
    {('pycodestyle', 'E701'), ('pylint', 'C0321')},

    # unused variable
    {('pylint', 'W0612'), ('pyflakes', 'W0612')},

    # undefined variable
    {('pylint', 'E0602'), ('pyflakes', 'E0602')},

    # unused import
    {('pylint', 'W0611'), ('pyflakes', 'W0611')},

    # whitespace before ')'
    {('pylint', 'C0326'), ('pycodestyle', 'E202')},

    # whitespace before '('
    {('pylint', 'C0326'), ('pycodestyle', 'E211')},

    # multiple spaces after operator
    {('pylint', 'C0326'), ('pycodestyle', 'E222')},

    # missing whitespace around operator
    {('pylint', 'C0326'), ('pycodestyle', 'E225')},

    # unexpected spaces
    {('pylint', 'C0326'), ('pycodestyle', 'E251')},

    # long lines
    {('pylint', 'C0301'), ('pycodestyle', 'E501')},

    # statement ends with a semicolon
    {('pylint', 'W0301'), ('pycodestyle', 'E703')},

    # multiple statements on one line
    {('pylint', 'C0321'), ('pycodestyle', 'E702')},

    # bad indentation
    {('pylint', 'W0311'), ('pycodestyle', 'E111')},

    # wildcart import
    {('pylint', 'W00401'), ('pyflakes', 'W0401')},

    # module docstring
    {('pydocstyle', 'D100'), ('pylint', 'C0111')},

) for key in values}


class Error:

    """ Store an error's information. """

    __slots__ = 'source', 'col', 'lnum', 'etype', 'message', 'filename', 'number'

    def __init__(self, source="pylama", col=1, lnum=1, type=None,
                 text="unknown error", filename="", number="", **_):
        """ Init error information with default values. """
        text = str(text).strip().replace('\n', ' ')
        if number:
            self.number = number
        else:
            number = PATTERN_NUMBER.match(text)
            self.number = number.group(1).upper() if number else ""

        self.etype = type[:1] if type else (number[0] if number else 'E')
        self.col = col
        self.filename = filename
        self.source = source
        self.lnum = int(lnum)
        self.message = text

    def __repr__(self):
        return f"<Error: {self.number} {self.message}>"

    def format(self, pattern: str) -> str:
        """Format the error with the given pattern."""
        return pattern.format(
            filename=self.filename,
            lnum=self.lnum,
            col=self.col,
            message=self.message,
            etype=self.etype,
            source=self.source,
            number=self.number,
        )

    def to_dict(self) -> Dict[str, Any]:
        """Return the error as a dict."""
        return {
            'source': self.source,
            'col': self.col,
            'lnum': self.lnum,
            'etype': self.etype,
            'message': self.message,
            'filename': self.filename,
            'number': self.number,
        }


def remove_duplicates(errors: List[Error]) -> Generator[Error, None, None]:
    """ Filter duplicates from given error's list. """
    passed: DefaultDict[str, Set] = defaultdict(set)
    for error in errors:
        key = error.source, error.number
        if key in DUPLICATES:
            if key in passed[error.lnum]:
                continue
            passed[error.lnum] = DUPLICATES[key]
        yield error


def default_sorter(err: Error) -> Any:
    """Sort by line number."""
    return err.lnum


# pylama:ignore=W0622,D,R0924



"""pycodestyle support."""
from pycodestyle import (BaseReport, StyleGuide, Checker, _parse_multi_options,
                         get_parser)

from pylama.lint import LinterV2 as Abstract
from pylama.context import RunContext


class Linter(Abstract):
    """pycodestyle runner."""

    name = "pycodestyle"

    def run_check(self, ctx: RunContext):  # noqa
        """Check code with pycodestyle."""
        params = ctx.get_params('pycodestyle')
        if params:
            parser = get_parser()
            for option in parser.option_list:
                if option.dest and option.dest in params:
                    value = params[option.dest]
                    if isinstance(value, str):
                        params[option.dest] = option.convert_value(option, value)

            for key in ["filename", "exclude", "select", "ignore"]:
                if key in params and isinstance(params[key], str):
                    params[key] = _parse_multi_options(params[key])

        style = StyleGuide(reporter=_PycodestyleReport, **params)
        style.options.report.ctx = ctx
        checker = Checker(ctx.filename, lines=ctx.lines, options=style.options)
        checker.check_all()


class _PycodestyleReport(BaseReport):

    ctx: RunContext

    def error(self, line_number, offset, text, _):
        """Save errors."""
        code, _,  text = text.partition(' ')
        self.ctx.push(
            text=text,
            type=code[0],
            number=code,
            col=offset + 1,
            lnum=line_number,
            source='pycodestyle',
        )



"""Pylint integration to Pylama."""
import logging
from os import environ
from pathlib import Path
from typing import Dict

from pylint.lint import Run
from pylint.reporters import BaseReporter

from pylama.context import RunContext
from pylama.lint import LinterV2 as BaseLinter

HOME_RCFILE = Path(environ.get("HOME", "")) / ".pylintrc"


logger = logging.getLogger("pylama")


class Linter(BaseLinter):
    """Check code with Pylint."""

    name = "pylint"

    def run_check(self, ctx: RunContext):
        """Pylint code checking."""
        logger.debug("Start pylint")
        params = ctx.get_params("pylint")

        class Reporter(BaseReporter):
            """Handle messages."""

            def _display(self, _):
                pass

            def handle_message(self, msg):
                msg_id = msg.msg_id
                ctx.push(
                    col=msg.column,
                    lnum=msg.line,
                    number=msg_id,
                    text=msg.msg,
                    type=msg_id[0],
                    source="pylint",
                )

        params = _Params(params)
        logger.debug(params)

        reporter = Reporter()
        Run([ctx.temp_filename] + params.to_attrs(), reporter=reporter, exit=False)


class _Params:
    """Store pylint params."""

    def __init__(self, params: Dict):
        attrs = {
            name.replace("_", "-"): self.prepare_value(value)
            for name, value in params.items()
            if value is not None
        }
        if HOME_RCFILE.exists():
            attrs["rcfile"] = HOME_RCFILE.as_posix()

        self.attrs = attrs

    @staticmethod
    def prepare_value(value):
        """Prepare value to pylint."""
        if isinstance(value, (list, tuple, set)):
            return ",".join(value)

        if isinstance(value, bool):
            return "y" if value else "n"

        return str(value)

    def to_attrs(self):
        """Convert to argument list."""
        return [f"--{key}={value}" for key, value in self.attrs.items()]  # noqa

    def __str__(self):
        return " ".join(self.to_attrs())

    def __repr__(self):
        return f"<Pylint {self}>"


# pylama:ignore=W0403



#!/usr/bin/env python

"""Setup pylama installation."""

import pathlib

import pkg_resources
from setuptools import setup


def parse_requirements(path: str) -> "list[str]":
    with pathlib.Path(path).open(encoding='utf-8') as requirements:
        return [str(req) for req in pkg_resources.parse_requirements(requirements)]


OPTIONAL_LINTERS = ['pylint', 'eradicate', 'radon', 'mypy', 'vulture']


setup(
    install_requires=parse_requirements("requirements/requirements.txt"),
    extras_require=dict(
        tests=parse_requirements("requirements/requirements-tests.txt"),
        all=OPTIONAL_LINTERS, **{linter: [linter] for linter in OPTIONAL_LINTERS}
    ),
)



"""Pylama's core functionality.

Prepare params, check a modeline and run the checkers.
"""
import os.path as op
from typing import List

from pylama.config import CURDIR, LOGGER, Namespace
from pylama.context import RunContext
from pylama.errors import Error, default_sorter, remove_duplicates
from pylama.lint import LINTERS, LinterV2


def run(
    path: str, code: str = None, rootdir: str = CURDIR, options: Namespace = None
) -> List[Error]:
    """Run code checkers with the given params.

    :param path: (str) A file's path.
    """
    path = op.relpath(path, rootdir)

    with RunContext(path, code, options) as ctx:
        if ctx.skip:
            LOGGER.info("Skip checking for path: %s", path)

        else:
            for lname in ctx.linters or LINTERS:
                linter = LINTERS.get(lname)()
                LOGGER.info("Run [%s] %s", lname, path)
                if isinstance(linter, LinterV2):
                    linter.run_check(ctx)
                else:
                    for err_info in linter.run(
                        ctx.temp_filename, code=ctx.source, params=ctx.get_params(lname)
                    ):
                        ctx.push(source=lname, **err_info)

    if not ctx.errors:
        return ctx.errors

    errors = list(remove_duplicates(ctx.errors))

    sorter = default_sorter
    if options and options.sort:
        sort = options.sort
        sorter = lambda err: (sort.get(err.etype, 999), err.lnum)

    return sorted(errors, key=sorter)


# pylama:ignore=R0912,D210,F0001
